/* UPS auto shutdown daemon upsd.c

  Wire Connection

  PC Serial port (DB-25)            UPS MUE1011-GL
   4 RTS ----+----+
             |    |  R=10kOhm
             |    R
             R    |          +-----7 GND OF UPS STOP
             |    |          |
   8 DCD ----+-------------------- 1 LOW BATTERY
                  |          |
   7 GND --------------------+---- 5 SIGNAL GROUND
                  |
   5 CTS ---------+--------------- 3 POWER FAIL

  20 DTR ------------------------- 6 UPS STOP
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <unistd.h>
#include <varargs.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/ttycom.h>
#include <sys/time.h>

#define	TURN_UPS_OFF	0	/* Yes : 1 , No : 0 */

int status;
char *scr = "./rc.shutdown";
char *idf = "./upsd.pid";

void upsterm();
void upsdown(int);

int main(int argc, char *argv[]) {
  FILE *fd;
  struct timeval tp;
  struct timezone tzp;
  long powerdowntime, standuptime=60;
  int pd, zero=0, low=0, pow=1, lineok, batteryok;

  /* check arguments */
  switch(argc) {
  case  4:
    scr = argv[3];
  case  3:
    standuptime = atol(argv[2]);
  case  2:
    break;
  default:
    fprintf(stderr, "usage: %s <device> [wait [script]]\n", argv[0]);
    exit(1);
  }

  /* check if script exists */
  if (!(fd = fopen(scr, "r"))) {
    fprintf(stderr, "fopen: %s: %s\n", scr, sys_errlist[errno]);
    exit(1);
  }
  fclose(fd);

  /* check if upsd is already running */
  if (fd = fopen(idf, "r")) {
    fprintf(stderr, "fopen: %s: File already exists\n", idf);
    exit(1);
  }

  /* become a daemon */
  switch(fork()) {
  case -1:
    /* error */
    fprintf(stderr, "fork: %s\n", sys_errlist[errno]);
    exit(1);
  case  0:
    /* child */
    break;
  default:
    /* parent */
    exit(0);
  }

  /* save the pid */
  if (!(fd = fopen(idf, "w"))) {
    fprintf(stderr, "fopen: %s: %s\n", idf, sys_errlist[errno]);
    exit(1);
  }
  fprintf(fd, "%d\n", (int)getpid());
  fclose(fd);

  /* open monitor device */
  if ((pd = open(argv[1], O_RDWR | O_NDELAY)) < 0) {
    fprintf(stderr, "open: %s: %s\n", argv[1], sys_errlist[errno]);
    exit(1);
  }

  /* daemon is alive */
  openlog("upsd", LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "daemon started");

  /* signal reaction */
  (void)signal(SIGTERM, upsterm);

  /* monitor device */
  while (1) {
    /* clear bits */
    if (ioctl(pd, TIOCMSET, &zero) < 0) {
      fprintf(stderr, "ioctl: %s\n", sys_errlist[errno]);
      exit(1);
    }

    /* set RTS monitor powerfail and low battery */
    status = TIOCM_RTS;
    if (ioctl(pd, TIOCMSET, &status) < 0) {
      fprintf(stderr, "ioctl: %s\n", sys_errlist[errno]);
      exit(1);
    }
    usleep(1000); /* wait for RTS stable */

    /* get device status */
    if (ioctl(pd, TIOCMGET, &status) < 0) {
      fprintf(stderr, "ioctl: %s\n", sys_errlist[errno]);
      exit(1);
    }

    /* determin status */
    lineok = status & TIOCM_CTS;
    batteryok = status & TIOCM_CD;
    printf("Status:[%d]batteryok[%d]linkok[%d]\n",status,batteryok,lineok);

    /* power down */
    if (!lineok) {
      if (pow) {
//	syslog(LOG_CRIT, "line failure");
	gettimeofday(&tp,&tzp);
	powerdowntime = tp.tv_sec;
	pow = 0;

	/* down ups */
	if (low) {
	  printf("!linkok & ( pow & low ) \n");
	  //	  upsdown(pd);
	}
      }
      else {
	gettimeofday(&tp,&tzp);
	if ((tp.tv_sec-powerdowntime) > standuptime) {
	  printf("!linkok & !pow & (tp.tv_sec-powerdowntime)>standuptime \n");
	  //	  upsdown(pd);
	}
      }

    /* power up */
    } else {
      if (!pow) {
//	syslog(LOG_CRIT, "line ok");
	pow = 1;
      }
    }

    /* battery ok */
    if (batteryok) {
      if (low) {
//	syslog(LOG_CRIT, "battery ok");
	low = 0;
      }

    /* battery low */
    } else {
      if (!low) {
//	syslog(LOG_ALERT, "battery low");
	low = 1;

	/* down ups */
	if (!pow) {
	  printf("!batteryok & !low & !pow\n");
//	  upsdown(pd);
	}
      }
    }

    sleep(1);
  }

  /* not reached */
  return 0;

}

void upsterm() {
  /* log termination message */
  syslog(LOG_INFO, "daemon terminated");

  /* remove pid file */
  unlink(idf);

  exit(0);
}

void upsdown(int pd) {
  /* log shutdown message */
  syslog(LOG_ALERT, "system is going down");

  /* remove pid file */
  unlink(idf);

  /* save our filesystem */
  system("/bin/sync");
  system("/bin/sync");
  system("/bin/sync");

#if TURN_UPS_OFF
  /* shutdown ups */
  status = TIOCM_DTR;
  if (ioctl(pd, TIOCMSET, &status) < 0) {
    fprintf(stderr, "ioctl: %s\n", sys_errlist[errno]);
    exit(1);
  }
  sleep(1);  /* wait shutdown signal period */
#endif

  /* run shutdown script */
  system(scr);
  exit(0); /* exit to keep DTR set */
}
